(**************************************************************************)
(*                                                                        *)
(*  Lucid Synchrone                                                       *)
(*                                                                        *)
(*  Author : Marc Pouzet                                                  *)
(*  Organization : Demons, LRI, University of Paris-Sud, Orsay            *)
(*                                                                        *)
(**************************************************************************)

(* $Id: chrono.ls,v 1.3 2007/01/02 12:13:33 pouzet Exp $ *)

(* this module provides the necessary stuff for emitting *)
(* input/outputs for the "sim2chro" chronogram tool *)
(* the chronogram is produced in a file [oc] *)

open List

type types = Tint | Tbool | Tfloat
type tval = Vint of int | Vbool of bool | Vfloat of float

let output_type oc typ =
  match typ with
      Tint -> output_string oc "int"
    | Tbool -> output_string oc "bool"
    | Tfloat -> output_string oc "float"
  end

let output_init1 oc is_input (name, typ) =
  output_string oc "\"";
  output_string oc is_input;
  output_string oc " ";
  output_string oc name;
  output_string oc "\":";
  output_type oc typ;
  output_string oc "\n"
    

let output_init oc p def_inputs def_outputs =
  output_string oc ("#program \"" ^ p ^ "\"");
  output_string oc "\n";
  output_string oc "#@inputs\n";
  List.iter (output_init1 oc "Input") def_inputs;
  output_string oc "@#";
  output_string oc "#@outputs\n";
  List.iter (output_init1 oc "Output") def_outputs;
  output_string oc "@#\n";
  flush oc

let output_data oc v =
  match v with
      Vint(v) -> output_string oc (string_of_int v)
    | Vbool(b) -> output_string oc (if b then "t" else "f")
    | Vfloat(f) -> output_string oc (string_of_float f)
  end

let output_step oc step (i, o) =
  let print_value v =
    output_data oc v;
    output_string oc "\t" in
  output_string oc "#step ";
  output_string oc (string_of_int step);
  output_string oc "\n"; flush oc;
  List.iter print_value i;
  output_string oc "#outs\t";
  List.iter print_value o;
  output_string oc "\n";
  flush oc

(* emitting input/outputs for Sim2Chro *)
let sim2chro oc =
  (* initialisation code *)
  let initialisation p def_inputs def_outputs =
    output_init oc p def_inputs def_outputs in
    
  (* step code *)
  let node running (inputs, outputs) =
    let rec step = 1 -> pre step + 1 in
    output_step oc step (inputs, outputs) in
    
  initialisation, running

(*let node main () =
  let initialisation, running = sim2chro stdout in
  match true -> false with
      true ->
	initialisation "Main"
	  ["one", Tint; "half", Tbool] 
	  ["open_light", Tbool; "open_gas", Tbool]
    | false -> ()
  end;
  let rec nat = 0 -> (pre nat + 1) mod 10 in
  let rec half = true fby true fby not half in
  running ([Vint(nat); Vbool(half)], [Vbool(true); Vbool(true)])*)
