;; lucy.el --- Lucid Synchrone code editing commands for Emacs

;;indentation code is Copyright (C) 1996 by Ian T Zimmerman <itz@rahul.net>
;;copying: covered by the current FSF General Public License.

;; indentation code adapted for Objective Caml by Jacques Garrigue,
;; july 1997. <garrigue@kurims.kyoto-u.ac.jp>

;; indentation code adapted for Lucid Synchrone by Sarah Maarek,
;; july 2004. 

;;user customizable variables
(defvar lucy-quote-char "'"
  "*Quote for character constants. \"'\" for Lucid Synchrone, \"`\" for Caml-Light.")

(defvar lucy-imenu-enable nil
  "*Enable Imenu support.")

(defvar lucy-mode-indentation 2
  "*Used for \\[lucy-unindent-command].")

(defvar lucy-lookback-limit 5000
  "*How far to look back for syntax things in lucy mode.")

(defvar lucy-max-indent-priority 8
  "*Bounds priority of operators permitted to affect lucy indentation.

Priorities are assigned to `interesting' lucy operators as follows:

        all keywords 0 to 7     8
        type, val, ... + 0      7
        :: ^                    6
        @                       5
        := <-                   4
        if                      3
        fun, let, match ...     2
        module                  1
        opening keywords        0.")

(defvar lucy-apply-extra-indent 2
  "*How many spaces to add to indentation for an application in lucy mode.")
(make-variable-buffer-local 'lucy-apply-extra-indent)

(defvar lucy-automaton-indent 2
  "*How many spaces to indent from a automaton keyword in lucy mode.")
(make-variable-buffer-local 'lucy-automaton-indent)

(defvar lucy-loop-indent 2
  "*How many spaces to indent from a loop keyword in lucy mode.")
(make-variable-buffer-local 'lucy-loop-indent)

(defvar lucy-class-indent 2
  "*How many spaces to indent from a class keyword in lucy mode.")
(make-variable-buffer-local 'lucy-class-indent)

(defvar lucy-control-indent 2
  "*How many spaces to indent from a control keyword in lucy mode.")
(make-variable-buffer-local 'lucy-control-indent)

(defvar lucy-do-indent 2
  "*How many spaces to indent from a do keyword in lucy mode.")
(make-variable-buffer-local 'lucy-do-indent)

(defvar lucy-reset-indent 2
  "*How many spaces to indent from a reset keyword in lucy mode.")
(make-variable-buffer-local 'lucy-reset-indent)

(defvar lucy-exception-indent 2
  "*How many spaces to indent from a exception keyword in lucy mode.")
(make-variable-buffer-local 'lucy-exception-indent)

(defvar lucy-for-indent 2
  "*How many spaces to indent from a for keyword in lucy mode.")
(make-variable-buffer-local 'lucy-for-indent)

(defvar lucy-fun-indent 2
  "*How many spaces to indent from a fun keyword in lucy mode.")
(make-variable-buffer-local 'lucy-fun-indent)

(defvar lucy-function-indent 4
  "*How many spaces to indent from a function keyword in lucy mode.")
(make-variable-buffer-local 'lucy-function-indent)

(defvar lucy-if-indent  2
  "*How many spaces to indent from a if keyword in lucy mode.")
(make-variable-buffer-local 'lucy-if-indent)

(defvar lucy-if-else-indent 0
  "*How many spaces to indent from an if .. else line in lucy mode.")
(make-variable-buffer-local 'lucy-if-else-indent)

(defvar lucy-inherit-indent 2
  "*How many spaces to indent from a inherit keyword in lucy mode.")
(make-variable-buffer-local 'lucy-inherit-indent)

(defvar lucy-initializer-indent 2
  "*How many spaces to indent from a initializer keyword in lucy mode.")
(make-variable-buffer-local 'lucy-initializer-indent)

(defvar lucy-include-indent 2
  "*How many spaces to indent from a include keyword in lucy mode.")
(make-variable-buffer-local 'lucy-include-indent)

(defvar lucy-let-indent 2
  "*How many spaces to indent from a let keyword in lucy mode.")
(make-variable-buffer-local 'lucy-let-indent)

(defvar lucy-rec-indent 2
  "*How many spaces to indent from a rec keyword in lucy mode.")
(make-variable-buffer-local 'lucy-rec-indent)

(defvar lucy-let-in-indent 0
  "*How many spaces to indent from a let .. in keyword in lucy mode.")
(make-variable-buffer-local 'lucy-let-in-indent)

(defvar lucy-let-where-indent 0
  "*How many spaces to indent from a let .. where keyword in lucy mode.")
(make-variable-buffer-local 'lucy-let-where-indent)

(defvar lucy-match-indent 2
  "*How many spaces to indent from a match keyword in lucy mode.")
(make-variable-buffer-local 'lucy-match-indent)

(defvar lucy-method-indent 2
  "*How many spaces to indent from a method keyword in lucy mode.")
(make-variable-buffer-local 'lucy-method-indent)

(defvar lucy-module-indent 2
  "*How many spaces to indent from a module keyword in lucy mode.")
(make-variable-buffer-local 'lucy-module-indent)

(defvar lucy-object-indent 2
  "*How many spaces to indent from a object keyword in lucy mode.")
(make-variable-buffer-local 'lucy-object-indent)

(defvar lucy-of-indent 2
  "*How many spaces to indent from a of keyword in lucy mode.")
(make-variable-buffer-local 'lucy-of-indent)

(defvar lucy-parser-indent 4
  "*How many spaces to indent from a parser keyword in lucy mode.")
(make-variable-buffer-local 'lucy-parser-indent)

(defvar lucy-sig-indent 2
  "*How many spaces to indent from a sig keyword in lucy mode.")
(make-variable-buffer-local 'lucy-sig-indent)

(defvar lucy-struct-indent 2
  "*How many spaces to indent from a struct keyword in lucy mode.")
(make-variable-buffer-local 'lucy-struct-indent)

(defvar lucy-try-indent 2
  "*How many spaces to indent from a try keyword in lucy mode.")
(make-variable-buffer-local 'lucy-try-indent)

(defvar lucy-type-indent 4
  "*How many spaces to indent from a type keyword in lucy mode.")
(make-variable-buffer-local 'lucy-type-indent)

(defvar lucy-val-indent 2
  "*How many spaces to indent from a val keyword in lucy mode.")
(make-variable-buffer-local 'lucy-val-indent)

(defvar lucy-::-indent  2
  "*How many spaces to indent from a :: operator in lucy mode.")
(make-variable-buffer-local 'lucy-::-indent)

(defvar lucy-@-indent   2
  "*How many spaces to indent from a @ operator in lucy mode.")
(make-variable-buffer-local 'lucy-@-indent)

(defvar lucy-:=-indent  2
  "*How many spaces to indent from a := operator in lucy mode.")
(make-variable-buffer-local 'lucy-:=-indent)

(defvar lucy-<--indent  2
  "*How many spaces to indent from a <- operator in lucy mode.")
(make-variable-buffer-local 'lucy-<--indent)

(defvar lucy-->-indent  2
  "*How many spaces to indent from a -> operator in lucy mode.")
(make-variable-buffer-local 'lucy-->-indent)

(defvar lucy-lb-indent 2
  "*How many spaces to indent from a \[ operator in lucy mode.")
(make-variable-buffer-local 'lucy-lb-indent)

(defvar lucy-lc-indent 2
  "*How many spaces to indent from a \{ operator in lucy mode.")
(make-variable-buffer-local 'lucy-lc-indent)

(defvar lucy-lp-indent  1
  "*How many spaces to indent from a \( operator in lucy mode.")
(make-variable-buffer-local 'lucy-lp-indent)

(defvar lucy-and-extra-indent nil
  "*Extra indent for lucy lines starting with the and keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-and-extra-indent)

(defvar lucy-do-extra-indent nil
  "*Extra indent for lucy lines starting with the do keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-do-extra-indent)

(defvar lucy-done-extra-indent nil
  "*Extra indent for lucy lines starting with the done keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-done-extra-indent)

(defvar lucy-else-extra-indent nil
  "*Extra indent for lucy lines starting with the else keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-else-extra-indent)

(defvar lucy-end-extra-indent nil
  "*Extra indent for lucy lines starting with the end keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-end-extra-indent)

(defvar lucy-in-extra-indent nil
  "*Extra indent for lucy lines starting with the in keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-in-extra-indent)

(defvar lucy-where-extra-indent nil
  "*Extra indent for lucy lines starting with the where keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-where-extra-indent)

(defvar lucy-then-extra-indent nil
  "*Extra indent for lucy lines starting with the then keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-then-extra-indent)

(defvar lucy-continue-extra-indent nil
  "*Extra indent for lucy lines starting with the continue keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-continue-extra-indent)

(defvar lucy-to-extra-indent -1
  "*Extra indent for lucy lines starting with the to keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-to-extra-indent)

(defvar lucy-until-extra-indent nil
  "*Extra indent for lucy lines starting with the until keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-until-extra-indent)

(defvar lucy-every-extra-indent nil
  "*Extra indent for lucy lines starting with the every keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-every-extra-indent)

(defvar lucy-unless-extra-indent nil
  "*Extra indent for lucy lines starting with the unless keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-unless-extra-indent)

(defvar lucy-with-extra-indent nil
  "*Extra indent for lucy lines starting with the with keyword.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-with-extra-indent)

(defvar lucy-comment-indent 3
  "*Indent inside comments.")
(make-variable-buffer-local 'lucy-comment-indent)

(defvar lucy-|-extra-indent -2
  "*Extra indent for lucy lines starting with the | operator.
Usually negative. nil is align on master.")
(make-variable-buffer-local 'lucy-|-extra-indent)

(defvar lucy-rb-extra-indent -2
  "*Extra indent for lucy lines statring with ].
Usually negative. nil is align on master.")

(defvar lucy-rc-extra-indent -2
  "*Extra indent for lucy lines starting with }.
Usually negative. nil is align on master.")

(defvar lucy-rp-extra-indent -1
  "*Extra indent for lucy lines starting with ).
Usually negative. nil is align on master.")

(defvar lucy-electric-indent t
  "*Non-nil means electrically indent lines starting with |, ] or }.

Many people find eletric keys irritating, so you can disable them if
you are one.")

(defvar lucy-electric-close-vector t
  "*Non-nil means electrically insert a | before a vector-closing ].

Many people find eletric keys irritating, so you can disable them if
you are one. You should probably have this on, though, if you also
have lucy-electric-indent on, which see.")

;;code
(if (or (not (fboundp 'indent-line-to))
        (not (fboundp 'buffer-substring-no-properties)))
    (require 'lucy-compat))

(defvar lucy-shell-active nil
  "Non nil when a subshell is running.")

(defvar running-xemacs  (string-match "XEmacs" emacs-version)
  "Non-nil if we are running in the XEmacs environment.")

(defvar lucy-mode-map nil
  "Keymap used in Lucy mode.")
(if lucy-mode-map
    ()
  (setq lucy-mode-map (make-sparse-keymap))
  (define-key lucy-mode-map "|" 'lucy-electric-pipe)
  (define-key lucy-mode-map "}" 'lucy-electric-pipe)
  (define-key lucy-mode-map "]" 'lucy-electric-rb)
  (define-key lucy-mode-map "\t" 'lucy-indent-command)
  (define-key lucy-mode-map [backtab] 'lucy-unindent-command)

;itz 04-21-96 instead of defining a new function, use defadvice
;that way we get out effect even when we do \C-x` in compilation buffer
;  (define-key lucy-mode-map "\C-x`" 'lucy-next-error)

  (if running-xemacs
      (define-key lucy-mode-map 'backspace 'backward-delete-char-untabify)
    (define-key lucy-mode-map "\177" 'backward-delete-char-untabify))

  ;; lucy-types
  (define-key lucy-mode-map [?\C-c?\C-t] 'lucy-types-show-type)
  ;; to prevent misbehavior in case of error during exploration.
  (define-key lucy-mode-map [mouse-2] 'lucy-types-mouse-ignore)
  (define-key lucy-mode-map [down-mouse-2] 'lucy-types-explore)
  ;; lucy-help
  (define-key lucy-mode-map [?\C-c?i] 'lucy-add-path)
  (define-key lucy-mode-map [?\C-c?]] 'lucy-close-module)
  (define-key lucy-mode-map [?\C-c?[] 'lucy-open-module)
  (define-key lucy-mode-map [?\C-c?\C-h] 'lucy-help)
  (define-key lucy-mode-map [?\C-c?\t] 'lucy-complete)
  ;; others
  (define-key lucy-mode-map "\C-ca" 'lucy-insert-automaton-form)
  (define-key lucy-mode-map "\C-cf" 'lucy-insert-for-form)
  (define-key lucy-mode-map "\C-cc" 'lucy-insert-control-form)
  (define-key lucy-mode-map "\C-co" 'lucy-insert-loop-form)
  (define-key lucy-mode-map "\C-cr" 'lucy-insert-reset-form)
  (define-key lucy-mode-map "\C-cw" 'lucy-insert-do-until-form)
  (define-key lucy-mode-map "\C-cs" 'lucy-insert-do-unless-form)  
  (define-key lucy-mode-map "\C-ci" 'lucy-insert-if-form)
  (define-key lucy-mode-map "\C-cl" 'lucy-insert-let-form)
  (define-key lucy-mode-map "\C-cm" 'lucy-insert-match-form)
  (define-key lucy-mode-map "\C-ct" 'lucy-insert-try-form)
  (define-key lucy-mode-map "\C-c`" 'lucy-goto-phrase-error)
  (define-key lucy-mode-map "\C-c\C-a" 'lucy-find-alternate-file)
  (define-key lucy-mode-map "\C-c\C-c" 'compile)
  (define-key lucy-mode-map "\C-c\C-e" 'lucy-eval-phrase)
  (define-key lucy-mode-map "\C-c\C-\[" 'lucy-backward-to-less-indent)
  (define-key lucy-mode-map "\C-c\C-\]" 'lucy-forward-to-less-indent)
  (define-key lucy-mode-map "\C-c\C-q" 'lucy-indent-phrase)
  (define-key lucy-mode-map "\C-c\C-r" 'lucy-eval-region)
  (define-key lucy-mode-map "\C-c\C-s" 'lucy-show-subshell)
  (define-key lucy-mode-map "\M-\C-h" 'lucy-mark-phrase)
  (define-key lucy-mode-map "\M-\C-q" 'lucy-indent-phrase)
  (define-key lucy-mode-map "\M-\C-x" 'lucy-eval-phrase)

  (if running-xemacs nil ; if not running xemacs
    (let ((map (make-sparse-keymap "Lucy"))
          (forms (make-sparse-keymap "Forms")))
      (define-key lucy-mode-map "\C-c\C-d" 'lucy-show-imenu)
      (define-key lucy-mode-map [menu-bar] (make-sparse-keymap))
      (define-key lucy-mode-map [menu-bar lucy] (cons "Lucid Synchrone" map))
      ;; lucy-help

      (define-key map [open] '("Open add path" . lucy-add-path ))
      (define-key map [close]
         '("Close module for help" . lucy-close-module))
      (define-key map [open] '("Open module for help" . lucy-open-module))
      (define-key map [help] '("Help for identifier" . lucy-help))
      (define-key map [complete] '("Complete identifier" . lucy-complete))
      (define-key map [separator-help] '("---"))

      ;; lucy-types
      (define-key map [show-type]
          '("Show type at point" . lucy-types-show-type ))
      (define-key map [separator-types] '("---"))

      ;; others
      (define-key map [run-lucy] '("Start subshell..." . run-lucy))
      (define-key map [compile] '("Compile..." . compile))
      (define-key map [switch-view]
        '("Switch view" . lucy-find-alternate-file))
      (define-key map [separator-format] '("--"))
      (define-key map [forms] (cons "Forms" forms))
      (define-key map [show-imenu] '("Show index" . lucy-show-imenu))
      (put 'lucy-show-imenu 'menu-enable '(not lucy-imenu-shown))
      (define-key map [show-subshell] '("Show subshell" . lucy-show-subshell))
      (put 'lucy-show-subshell 'menu-enable 'lucy-shell-active)
      (define-key map [eval-phrase] '("Eval phrase" . lucy-eval-phrase))
      (put 'lucy-eval-phrase 'menu-enable 'lucy-shell-active)
      (define-key map [indent-phrase] '("Indent phrase" . lucy-indent-phrase))
      (define-key forms [control] '("control .. with .." . lucy-insert-control-form))
      (define-key forms [reset] '("reset .. every .." . lucy-insert-reset-form))
      (define-key forms [try] '("try .. with .." . lucy-insert-try-form))
      (define-key forms [until] '("do .. until .. then .." . lucy-insert-do-until-form))
      (define-key forms [unless] '("do .. unless .. then .." . lucy-insert-do-unless-form))
      (define-key forms [match] '("match .. with .." . lucy-insert-match-form))
      (define-key forms [let] '("let .. in .." . lucy-insert-let-form))
      (define-key forms [if] '("if .. then .. else .." . lucy-insert-if-form))
      (define-key forms [for] '("for .. do .. end" . lucy-insert-for-form))
      (define-key forms [automaton] '("automaton .. end" . lucy-insert-automaton-form))
      (define-key forms [loop] '("loop .. end" . lucy-insert-loop-form)))))

(defvar lucy-mode-xemacs-menu
  (if running-xemacs
      '("Lucid Synchrone"
        [ "Indent phrase" lucy-indent-phrase :keys "C-M-q" ]
        [ "Eval phrase" lucy-eval-phrase
          :active lucy-shell-active :keys "C-M-x" ]
        [ "Show subshell" lucy-show-subshell lucy-shell-active ]
        ("Forms"
         [ "control .. with .." lucy-insert-control-form t ]
	 [ "reset .. every .." lucy-insert-reset-form t ]
	 [ "try .. with .." lucy-insert-try-form t ]
         [ "do .. until .. then .." lucy-insert-do-until-form t ]
         [ "do .. unless .. then .." lucy-insert-do-unless-form t ]
         [ "match .. with .." lucy-insert-match-form t ]         
         [ "let .. in .." lucy-insert-let-form t ]
         [ "if .. then .. else .." lucy-insert-if-form t ]
         [ "for .. do .. end" lucy-insert-for-form t ]
	 [ "automaton .. end" lucy-insert-automaton-form t ]
	 [ "loop .. end" lucy-insert-loop-form t ])
        "---"
        [ "Switch view" lucy-find-alternate-file t ]
        [ "Compile..." compile t ]
        [ "Start subshell..." run-lucy t ]
        "---"
        [ "Show type at point" lucy-types-show-type t ]
        "---"
        [ "Complete identifier" lucy-complete t ]
        [ "Help for identifier" lucy-help t ]
        [ "Add path for documentation" lucy-add-path t ]
        [ "Open module for documentation" lucy-open t ]
        [ "Close module for documentation" lucy-close t ]
        ))
  "Menu to add to the menubar when running Xemacs")

(defvar lucy-mode-syntax-table nil
  "Syntax table in use in Lucy mode buffers.")
(if lucy-mode-syntax-table
    ()
  (setq lucy-mode-syntax-table (make-syntax-table))
  ; backslash is an escape sequence
  (modify-syntax-entry ?\\ "\\" lucy-mode-syntax-table)
  ; ( is first character of comment start
  (modify-syntax-entry ?\( "()1" lucy-mode-syntax-table)
  ; * is second character of comment start,
  ; and first character of comment end
  (modify-syntax-entry ?*  ". 23" lucy-mode-syntax-table)
  ; ) is last character of comment end
  (modify-syntax-entry ?\) ")(4" lucy-mode-syntax-table)
  ; backquote was a string-like delimiter (for character literals)
  ; (modify-syntax-entry ?` "\"" lucy-mode-syntax-table)
  ; quote and underscore are part of words
  (modify-syntax-entry ?' "w" lucy-mode-syntax-table)
  (modify-syntax-entry ?_ "w" lucy-mode-syntax-table)
  ; ISO-latin accented letters and EUC kanjis are part of words
  (let ((i 160))
    (while (< i 256)
      (modify-syntax-entry i "w" lucy-mode-syntax-table)
      (setq i (1+ i)))))

(defvar lucy-mode-abbrev-table nil
  "Abbrev table used for Lucy mode buffers.")
(if lucy-mode-abbrev-table nil
  (setq lucy-mode-abbrev-table (make-abbrev-table))
  (define-abbrev lucy-mode-abbrev-table "and" "and" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "do" "do" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "done" "done" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "else" "else" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "end" "end" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "every" "every" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "in" "in" 'lucy-abbrev-hook)
;  (define-abbrev lucy-mode-abbrev-table "rec" "rec" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "then" "then" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "continue" 
    "continue" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "until" "until" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "unless" "unless" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "when" "when" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "where" "where" 'lucy-abbrev-hook)
  (define-abbrev lucy-mode-abbrev-table "with" "with" 'lucy-abbrev-hook))

;; Other internal variables

(defvar lucy-last-noncomment-pos nil
  "Caches last buffer position determined not inside a lucy comment.")
(make-variable-buffer-local 'lucy-last-noncomment-pos)

;;last-noncomment-pos can be a simple position, because we nil it
;;anyway whenever buffer changes upstream. last-comment-start and -end
;;have to be markers, because we preserve them when the changes' end
;;doesn't overlap with the comment's start.

(defvar lucy-last-comment-start nil
  "A marker caching last determined lucy comment start.")
(make-variable-buffer-local 'lucy-last-comment-start)

(defvar lucy-last-comment-end nil
  "A marker caching last determined lucy comment end.")
(make-variable-buffer-local 'lucy-last-comment-end)

(make-variable-buffer-local 'before-change-function)

(defvar lucy-imenu-shown nil
  "True if we have computed definition list.")
(make-variable-buffer-local 'lucy-imenu-shown)

(defconst lucy-imenu-search-regexp
  (concat "\\<in\\>\\|"
          "^[ \t]*\\(let\\|class\\|type\\|m\\(odule\\|ethod\\)"
          "\\|functor\\|and\\|val\\)[ \t]+"
          "\\(\\('[a-zA-Z0-9]+\\|([^)]+)"
          "\\|mutable\\|node\\|private\\|rec\\|type\\)[ \t]+\\)?"
          "\\([a-zA-Z][a-zA-Z0-9_']*\\)"))

;;; The major mode
(eval-when-compile
  (if (and (boundp 'running-xemacs) running-xemacs) nil
    (require 'imenu)))

;;
(defvar lucy-mode-hook nil
  "Hook for lucy-mode")

(defun lucy-mode ()
  "Major mode for editing Lucid Synchrone code.

\\{lucy-mode-map}"

  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'lucy-mode)
  (setq mode-name "Lucid Synchrone")
  (use-local-map lucy-mode-map)
  (set-syntax-table lucy-mode-syntax-table)
  (setq local-abbrev-table lucy-mode-abbrev-table)
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  (make-local-variable 'comment-start)
  (setq comment-start "(*")
  (make-local-variable 'comment-end)
  (setq comment-end "*)")
  (make-local-variable 'comment-column)
  (setq comment-column 40)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "(\\*+ *")
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments nil)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'lucy-indent-command)
  ;itz Fri Sep 25 13:23:49 PDT 1998
  (make-local-variable 'add-log-current-defun-function)
  (setq add-log-current-defun-function 'lucy-current-defun)
  ;itz 03-25-96
  (setq before-change-function 'lucy-before-change-function)
  (setq lucy-last-noncomment-pos nil)
  (setq lucy-last-comment-start (make-marker))
  (setq lucy-last-comment-end (make-marker))
  ;garrigue 27-11-96
  (setq case-fold-search nil)
  ;garrigue july 97
  (if running-xemacs ; from Xemacs lisp mode
      (if (and (featurep 'menubar)
               current-menubar)
          (progn
            ;; make a local copy of the menubar, so our modes don't
            ;; change the global menubar
            (set-buffer-menubar current-menubar)
            (add-submenu nil lucy-mode-xemacs-menu)))
    ;imenu support (not for Xemacs)
    (make-local-variable 'imenu-create-index-function)
    (setq imenu-create-index-function 'lucy-create-index-function)
    (make-local-variable 'imenu-generic-expression)
    (setq imenu-generic-expression lucy-imenu-search-regexp)
    (if (and lucy-imenu-enable (< (buffer-size) 10000))
        (lucy-show-imenu)))
  (run-hooks 'lucy-mode-hook))

(defun lucy-set-compile-command ()
  "Hook to set compile-command locally, unless there is a Makefile in the 
   current directory." 
  (interactive)
  (unless (or (null buffer-file-name)
              (file-exists-p "makefile")
              (file-exists-p "Makefile"))
    (let* ((filename (file-name-nondirectory buffer-file-name))
           (basename (file-name-sans-extension filename))
           (command nil))
      (cond
       ((string-match ".*\\.mli\$" filename)
        (setq command "ocamlc"))
       ((string-match ".*\\.ml\$" filename)
        (setq command "ocamlc") ; (concat "ocamlc -o " basename)
        )
       ((string-match ".*\\.mll\$" filename)
        (setq command "ocamllex"))
       ((string-match ".*\\.mll\$" filename)
        (setq command "ocamlyacc"))
       )
      (if command
          (progn 
            (make-local-variable 'compile-command)
            (setq compile-command (concat command " " filename))))
      )))

(add-hook 'lucy-mode-hook 'lucy-set-compile-command)

;;; Auxiliary function. Garrigue 96-11-01.

(defun lucy-find-alternate-file ()
  (interactive)
  (let ((name (buffer-file-name)))
    (if (string-match "^\\(.*\\)\\.\\(ml\\|mli\\)$" name)
        (find-file
         (concat
          (lucy-match-string 1 name)
          (if (string= "ml" (lucy-match-string 2 name)) ".mli" ".ml"))))))

;;; subshell support

(defun lucy-eval-region (start end)
  "Send the current region to the inferior Lucy process."
  (interactive"r")
  (require 'inf-lucy)
  (inferior-lucy-eval-region start end))

;; old version ---to be deleted later
; 
; (defun lucy-eval-phrase ()
;   "Send the current Lucy phrase to the inferior Lucy process."
;   (interactive)
;   (save-excursion
;     (let ((bounds (lucy-mark-phrase)))
;     (inferior-lucy-eval-region (car bounds) (cdr bounds)))))

(defun lucy-eval-phrase (arg &optional min max)
  "Send the phrase containing the point to the LUCY process.
With prefix-arg send as many phrases as its numeric value, 
If an error occurs during evalutaion, stop at this phrase and
repport the error. 

Return nil if noerror and position of error if any.

If arg's numeric value is zero or negative, evaluate the current phrase
or as many as prefix arg, ignoring evaluation errors. 
This allows to jump other erroneous phrases. 

Optional arguments min max defines a region within which the phrase
should lies."
  (interactive "p")
  (require 'inf-lucy)
  (inferior-lucy-eval-phrase arg min max))

(defun lucy-eval-buffer (arg)
  "Evaluate the buffer from the beginning to the phrase under the point.
With prefix arg, evaluate past the whole buffer, no stopping at
the current point."
  (interactive "p")
  (let ((here (point)) err)
    (goto-char (point-min))
    (setq err
          (lucy-eval-phrase 500 (point-min) (if arg (point-max) here)))
    (if err (set-mark err))
    (goto-char here)))

(defun lucy-show-subshell ()
  (interactive)
  (require 'inf-lucy)
  (inferior-lucy-show-subshell))


;;; Imenu support
(defun lucy-show-imenu ()
  (interactive)
  (require 'imenu)
  (switch-to-buffer (current-buffer))
  (imenu-add-to-menubar "Defs")
  (setq lucy-imenu-shown t))

(defun lucy-prev-index-position-function ()
  (let (found data)
    (while (and (setq found
                      (re-search-backward lucy-imenu-search-regexp nil 'move))
                (progn (setq data (match-data)) t)
                (or (lucy-in-literal-p)
                    (lucy-in-comment-p)
                    (if (looking-at "in") (lucy-find-in-match)))))
    (set-match-data data)
    found))
(defun lucy-create-index-function ()
  (let (value-alist
        type-alist
        class-alist
        method-alist
        module-alist
        and-alist
        all-alist
        menu-alist
        (prev-pos (point-max))
        index)
    (goto-char prev-pos)
    (imenu-progress-message prev-pos 0 t)
    ;; collect definitions
    (while (lucy-prev-index-position-function)
      (setq index (cons (lucy-match-string 5) (point)))
      (imenu-progress-message prev-pos nil t)
      (setq all-alist (cons index all-alist))
      (cond
       ((looking-at "[ \t]*and")
        (setq and-alist (cons index and-alist)))
       ((looking-at "[ \t]*rec")
        (setq and-alist (cons index and-alist)))
       ((looking-at "[ \t]*let")
        (setq value-alist (cons index (append and-alist value-alist)))
        (setq and-alist nil))
       ((looking-at "[ \t]*type")
        (setq type-alist (cons index (append and-alist type-alist)))
        (setq and-alist nil))
       ((looking-at "[ \t]*class")
        (setq class-alist (cons index (append and-alist class-alist)))
        (setq and-alist nil))
       ((looking-at "[ \t]*val")
        (setq value-alist (cons index value-alist)))
       ((looking-at "[ \t]*\\(module\\|functor\\)")
        (setq module-alist (cons index module-alist)))
       ((looking-at "[ \t]*method")
        (setq method-alist (cons index method-alist)))))
    ;; build menu
    (mapcar
     '(lambda (pair)
        (if (symbol-value (cdr pair))
            (setq menu-alist
                  (cons
                   (cons (car pair)
                         (sort (symbol-value (cdr pair)) 'imenu--sort-by-name))
                   menu-alist))))
     '(("Values" . value-alist)
       ("Types" . type-alist)
       ("Modules" . module-alist)
       ("Methods" . method-alist)
       ("Classes" . class-alist)))
    (if all-alist (setq menu-alist (cons (cons "Index" all-alist) menu-alist)))
    (imenu-progress-message prev-pos 100 t)
    menu-alist))

;;; Indentation stuff

(defun lucy-in-indentation ()
  "Tests whether all characters between beginning of line and point
are blanks."
  (save-excursion
    (skip-chars-backward " \t")
    (bolp)))

;;; The command
;;; Sorry, I didn't like the previous behaviour... Garrigue 96/11/01

(defun lucy-indent-command (&optional p)
  "Indent the current line in Lucy mode.

Compute new indentation based on lucy syntax. If prefixed, indent
the line all the way to where point is."

  (interactive "*p")
  (cond
   ((and p (> p 1)) (indent-line-to (current-column)))
   ((lucy-in-indentation) (indent-line-to (lucy-compute-final-indent)))
   (t (save-excursion
        (indent-line-to
         (lucy-compute-final-indent))))))

(defun lucy-unindent-command ()

  "Decrease indentation by one level in Lucy mode.

Works only if the point is at the beginning of an indented line
\(i.e. all characters between beginning of line and point are
blanks\).  Does nothing otherwise. The unindent size is given by the
variable lucy-mode-indentation."

  (interactive "*")
  (let* ((begline
          (save-excursion
            (beginning-of-line)
            (point)))
         (current-offset
          (- (point) begline)))
    (if (and (>= current-offset lucy-mode-indentation)
             (lucy-in-indentation))
        (backward-delete-char-untabify lucy-mode-indentation))))

;;;
;;; Error processing
;;;

;; Error positions are given in bytes, not in characters
;; This function switches to monobyte mode

(if (not (fboundp 'char-bytes))
    (defalias 'forward-byte 'forward-char)
  (defun lucy-char-bytes (ch)
    (let ((l (char-bytes ch)))
      (if (> l 1) (- l 1) l)))
  (defun forward-byte (count)
    (if (> count 0)
        (while (> count 0)
          (let ((char (char-after)))
            (if (null char)
                (setq count 0)
              (setq count (- count (lucy-char-bytes (char-after))))
              (forward-char))))
      (while (< count 0)
        (let ((char (char-after)))
          (if (null char)
              (setq count 0)
            (setq count (+ count (lucy-char-bytes (char-before))))
            (backward-char))))
    )))

(require 'compile)

;; In Emacs 19, the regexps in compilation-error-regexp-alist do not
;; match the error messages when the language is not English.
;; Hence we add a regexp.

(defconst lucy-error-regexp
  "^[A-\377]+ \"\\([^\"\n]+\\)\", [A-\377]+ \\([0-9]+\\)[-,:]"
  "Regular expression matching the error messages produced by ocamlc.")

(if (boundp 'compilation-error-regexp-alist)
    (or (assoc lucy-error-regexp
               compilation-error-regexp-alist)
        (setq compilation-error-regexp-alist
              (cons (list lucy-error-regexp 1 2)
               compilation-error-regexp-alist))))

;; A regexp to extract the range info

(defconst lucy-error-chars-regexp
  ".*, .*, [A-\377]+ \\([0-9]+\\)-\\([0-9]+\\):"
  "Regular expression extracting the character numbers
from an error message produced by ocamlc.")

;; Wrapper around next-error.

(defvar lucy-error-overlay nil)

;;itz 04-21-96 somebody didn't get the documetation for next-error
;;right. When the optional argument is a number n, it should move
;;forward n errors, not reparse.

;itz 04-21-96 instead of defining a new function, use defadvice
;that way we get our effect even when we do \C-x` in compilation buffer

(defadvice next-error (after lucy-next-error activate)
 "Reads the extra positional information provided by the Lucy compiler.

Puts the point and the mark exactly around the erroneous program
fragment. The erroneous fragment is also temporarily highlighted if
possible."

 (if (eq major-mode 'lucy-mode)
     (let (bol beg end)
       (save-excursion
         (set-buffer
          (if (boundp 'compilation-last-buffer)
              compilation-last-buffer   ;Emacs 19
            "*compilation*"))           ;Emacs 18
         (save-excursion
           (goto-char (window-point (get-buffer-window (current-buffer))))
           (if (looking-at lucy-error-chars-regexp)
               (setq beg
                     (string-to-int
                      (buffer-substring (match-beginning 1) (match-end 1)))
                     end
                     (string-to-int
                      (buffer-substring (match-beginning 2) (match-end 2)))))))
       (cond (beg
              (setq end (- end beg))
              (beginning-of-line)
              (forward-byte beg)
              (setq beg (point))
              (forward-byte end)
              (setq end (point))
              (goto-char beg)
              (push-mark end t)
              (cond ((fboundp 'make-overlay)
                     (if lucy-error-overlay ()
                       (setq lucy-error-overlay (make-overlay 1 1))
                       (overlay-put lucy-error-overlay 'face 'region))
                     (unwind-protect
                         (progn
                           (move-overlay lucy-error-overlay
                                         beg end (current-buffer))
                           (sit-for 60))
                       (delete-overlay lucy-error-overlay)))))))))

;; Usual match-string doesn't work properly with font-lock-mode
;; on some emacs.

(defun lucy-match-string (num &optional string)

  "Return string of text matched by last search, without properties.

NUM specifies which parenthesized expression in the last regexp.
Value is nil if NUMth pair didn't match, or there were less than NUM
pairs.  Zero means the entire text matched by the whole regexp or
whole string."

  (let* ((data (match-data))
         (begin (nth (* 2 num) data))
         (end (nth (1+ (* 2 num)) data)))
    (if string (substring string begin end)
      (buffer-substring-no-properties begin end))))

;; itz Thu Sep 24 19:02:42 PDT 1998 this is to have some level of
;; comfort when sending phrases to the toplevel and getting errors.
(defun lucy-goto-phrase-error ()
  "Find the error location in current Lucy phrase."
  (interactive)
  (require 'inf-lucy)
  (let ((bounds (save-excursion (lucy-mark-phrase))))
    (inferior-lucy-goto-error (car bounds) (cdr bounds))))

;;; Phrases

;itz the heuristics used to see if we're `between two phrases'
;didn't seem right to me.

(defconst lucy-phrase-start-keywords
  (concat "\\<\\(class\\|ex\\(ternal\\|ception\\)\\|functor"
          "\\|let\\|module\\|open\\|type\\|val\\)\\>")
  "Keywords starting phrases in files")

;; a phrase starts when a toplevel keyword is at the beginning of a line
(defun lucy-at-phrase-start-p ()
  (and (bolp)
       (or (looking-at "#")
           (looking-at lucy-phrase-start-keywords))))

(defun lucy-skip-comments-forward ()
  (skip-chars-forward " \n\t")
  (while (or (looking-at comment-start-skip) (lucy-in-comment-p))
    (if (= (following-char) ?\)) (forward-char)
      (search-forward comment-end))
    (skip-chars-forward " \n\t")))

(defun lucy-skip-comments-backward ()
  (skip-chars-backward " \n\t")
  (while (and (eq (preceding-char) ?\)) (eq (char-after (- (point) 2)) ?*))
    (backward-char)
    (while (lucy-in-comment-p) (search-backward comment-start))
    (skip-chars-backward " \n\t")))

(defconst lucy-phrase-sep-keywords (concat ";;\\|" lucy-phrase-start-keywords))

(defun lucy-find-phrase (&optional min-pos max-pos)
  "Find the LUCY phrase containing the point.
Return the position of the beginning of the phrase, and move point
to the end.
"
  (interactive)
  (if (not min-pos) (setq min-pos (point-min)))
  (if (not max-pos) (setq max-pos (point-max)))
  (let (beg end use-semi kwop)
    ;(lucy-skip-comments-backward)
    (cond
     ; shall we have special processing for semicolons?
     ;((and (eq (char-before (- (point) 1)) ?\;) (eq (char-before) ?\;))
     ; (forward-char)
     ; (lucy-skip-comments-forward)
     ; (setq beg (point))
     ; (while (and (search-forward ";;" max-pos 'move)
     ;    (or (lucy-in-comment-p) (lucy-in-literal-p)))))
     (t
      (lucy-skip-comments-forward)
      (if (lucy-at-phrase-start-p) (forward-char))
      (while (and (cond
                   ((re-search-forward lucy-phrase-sep-keywords max-pos 'move)
                    (goto-char (match-beginning 0)) t))
                  (or (not (or (bolp) (looking-at ";;")))
                      (lucy-in-comment-p)
                      (lucy-in-literal-p)))
        (forward-char))
      (setq end (+ (point) (if (looking-at ";;") 2 0)))
      (while (and
              (setq kwop (lucy-find-kwop lucy-phrase-sep-keywords min-pos))
              (not (string= kwop ";;"))
              (not (bolp))))
      (if (string= kwop ";;") (forward-char 2))
      (if (not kwop) (goto-char min-pos))
      (lucy-skip-comments-forward)
      (setq beg (point))
      (if (>= beg end) (error "no phrase before point"))
      (goto-char end)))
    (lucy-skip-comments-forward)
    beg))

(defun lucy-mark-phrase (&optional min-pos max-pos)
  "Put mark at end of this Lucy phrase, point at beginning.
"
  (interactive)
  (let* ((beg (lucy-find-phrase min-pos max-pos)) (end (point)))
    (push-mark)
    (goto-char beg)
    (cons beg end)))
    
;;itz Fri Sep 25 12:58:13 PDT 1998 support for adding change-log entries
(defun lucy-current-defun ()
  (save-excursion
    (lucy-mark-phrase)
    (if (not (looking-at lucy-phrase-start-keywords)) nil
      (re-search-forward lucy-phrase-start-keywords)
      (let ((done nil))
        (while (not done)
          (cond
           ((looking-at "\\s ")
            (skip-syntax-forward " "))
           ((char-equal (following-char) ?\( )
            (forward-sexp 1))
           ((char-equal (following-char) ?')
            (skip-syntax-forward "w_"))
           (t (setq done t)))))
      (re-search-forward "\\(\\sw\\|\\s_\\)+")
      (match-string 0))))

(defun lucy-overlap (b1 e1 b2 e2)
  (<= (max b1 b2) (min e1 e2)))

;this clears the last comment cache if necessary
(defun lucy-before-change-function (begin end)
  (if (and lucy-last-noncomment-pos
           (> lucy-last-noncomment-pos begin))
      (setq lucy-last-noncomment-pos nil))
  (if (and (marker-position lucy-last-comment-start)
           (marker-position lucy-last-comment-end)
           (lucy-overlap begin end
                         lucy-last-comment-start
                         lucy-last-comment-end))
      (prog2
          (set-marker lucy-last-comment-start nil)
          (set-marker lucy-last-comment-end nil)))
  (let ((orig-function (default-value 'before-change-function)))
    (if orig-function (funcall orig-function begin end))))

(defun lucy-in-literal-p ()
  "Returns non-nil if point is inside a lucy literal."
  (let* ((start-literal (concat "[\"" lucy-quote-char "]"))
         (char-literal
          (concat "\\([^\\]\\|\\\\\\.\\|\\\\[0-9][0-9][0-9]\\)"
                  lucy-quote-char))
         (pos (point))
         (eol (progn (end-of-line 1) (point)))
         state in-str)
    (beginning-of-line 1)
    (while (and (not state)
                (re-search-forward start-literal eol t)
                (<= (point) pos))
      (cond
       ((string= (lucy-match-string 0) "\"")
        (setq in-str t)
        (while (and in-str (not state)
                    (re-search-forward "\"\\|\\\\\"" eol t))
          (if (> (point) pos) (setq state t))
          (if (string= (lucy-match-string 0) "\"") (setq in-str nil)))
        (if in-str (setq state t)))
       ((looking-at char-literal)
        (if (and (>= pos (match-beginning 0)) (< pos (match-end 0)))
            (setq state t)
          (goto-char (match-end 0))))))
    (goto-char pos)
    state))

(defun lucy-forward-comment ()
  "Skip one (eventually nested) comment."
  (let ((count 1) match)
    (while (> count 0)
      (if (not (re-search-forward "(\\*\\|\\*)" nil 'move))
          (setq count -1)
        (setq match (lucy-match-string 0))
        (cond
         ((lucy-in-literal-p)
          nil)
         ((string= match comment-start)
          (setq count (1+ count)))
         (t
          (setq count (1- count))))))
    (= count 0)))

(defun lucy-backward-comment ()
  "Skip one (eventually nested) comment."
  (let ((count 1) match)
    (while (> count 0)
      (if (not (re-search-backward "(\\*\\|\\*)" nil 'move))
          (setq count -1)
        (setq match (lucy-match-string 0))
        (cond
         ((lucy-in-literal-p)
          nil)
         ((string= match comment-start)
          (setq count (1- count)))
         (t
          (setq count (1+ count))))))
    (= count 0)))

(defun lucy-in-comment-p ()
  "Returns non-nil if point is inside a lucy comment.
Returns nil for the parenthesis openning a comment."
  ;;we look for comments differently than literals. there are two
  ;;reasons for this. first, lucy has nested comments and it is not so
  ;;clear that parse-partial-sexp supports them; second, if proper
  ;;style is used, literals are never split across lines, so we don't
  ;;have to worry about bogus phrase breaks inside literals, while we
  ;;have to account for that possibility in comments.
  (if lucy-last-comment-start
      (save-excursion
        (let* ((cached-pos lucy-last-noncomment-pos)
               (cached-begin (marker-position lucy-last-comment-start))
               (cached-end (marker-position lucy-last-comment-end)))
          (cond
           ((and cached-begin cached-end
                 (< cached-begin (point)) (< (point) cached-end)) t)
           ((and cached-pos (= cached-pos (point))) nil)
           ((and cached-pos (> cached-pos (point))
                 (< (abs (- cached-pos (point))) lucy-lookback-limit))
            (let (end found (here (point)))
                                        ; go back to somewhere sure
              (goto-char cached-pos)
              (while (> (point) here)
                                        ; look for the end of a comment
                (while (and (if (search-backward comment-end (1- here) 'move)
                                (setq end (match-end 0))
                              (setq end nil))
                            (lucy-in-literal-p)))
                (if end (setq found (lucy-backward-comment))))
              (if (and found (= (point) here)) (setq end nil))
              (if (not end)
                  (setq lucy-last-noncomment-pos here)
                (set-marker lucy-last-comment-start (point))
                (set-marker lucy-last-comment-end end))
              end))
           (t
            (let (begin found (here (point)))
            ;; go back to somewhere sure (or far enough)
              (goto-char
               (if cached-pos cached-pos (- (point) lucy-lookback-limit)))
              (while (< (point) here)
                ;; look for the beginning of a comment
                (while (and (if (search-forward comment-start (1+ here) 'move)
                                (setq begin (match-beginning 0))
                              (setq begin nil))
                            (lucy-in-literal-p)))
                (if begin (setq found (lucy-forward-comment))))
              (if (and found (= (point) here)) (setq begin nil))
              (if (not begin)
                  (setq lucy-last-noncomment-pos here)
                (set-marker lucy-last-comment-start begin)
                (set-marker lucy-last-comment-end (point)))
              begin)))))))

;; Various constants and regexps

(defconst lucy-before-expr-prefix
  (concat "\\<\\(a\\(sr\\|utomaton\\)\\|c\\(lass\\|ont\\(inue\\|rol\\)\\)"
	  "\\|do\\(wnto\\)?\\|e\\(lse\\|very\\)"
          "\\|i\\(f\\|n\\(herit\\|itializer\\)?\\)"
	  "\\|f\\(or\\|un\\(ct\\(ion\\|or\\)\\)?\\)"          
          "\\|l\\(and\\|o\\(op\\|r\\)\\|s[lr]\\|xor\\)\\|m\\(atch\\|od\\)"
          "\\|o[fr]\\|parser\\|reset\\|s\\(ig\\|t\\(ate\\|ruct\\)\\)"
	  "\\|t\\(hen\\|o\\|ry\\)\\|un\\(less\\|til\\)"
	  "\\|w\\(hen\\|ith\\)\\)\\>\\>"
          "\\|[=<>@^|&+-*/$%][!$%*+-./:<=>?@^|~]*\\|:[:=]\\|[[({,;]")

  "Keywords that may appear immediately before an expression.
Used to distinguish it from toplevel let construct.")

(defconst lucy-matching-kw-regexp
  (concat
   "\\<\\(and\\|continue||\do\\(ne\\)?\\|e\\(lse\\|nd\\|very\\)"
   "\\|in\\|t\\(hen\\|o\\)\\|un\\(less\\|til\\)"
   "\\|w\\(he\\(n\\|re\\)\\|ith\\)\\)\\>\\|[^[|]|[^|]")
  "Regexp used in lucy mode for skipping back over nested blocks.")

(defconst lucy-matching-kw-alist
  '(("|" . lucy-find-pipe-match)
    (";" . lucy-find-semi-match)
    ("," . lucy-find-comma-match)
    ("continue" . lucy-find-continue-match)
    ("end" . lucy-find-end-match)
    ("every"  . lucy-find-every-match)
    ("done" . lucy-find-done-match)
    ("in"  . lucy-find-in-match)
    ("until"  . lucy-find-un-match)
    ("unless"  . lucy-find-un-match)
    ("when" . lucy-find-when-match)
    ("where" . lucy-find-where-match)
    ("with" . lucy-find-with-match)
    ("else" . lucy-find-else-match)
    ("then" . lucy-find-then-match)
    ("to" . lucy-find-end-match)
    ("do" . lucy-find-end-match)
    ("and" . lucy-find-and-match))

  "Association list used in lucy mode for skipping back over nested blocks.")

(defconst lucy-kwop-regexps (make-vector 9 nil)
  "Array of regexps representing lucy keywords of different priorities.")

(defun lucy-in-expr-p ()
  (let ((pos (point)) (in-expr t))
    (lucy-find-kwop
     (concat lucy-before-expr-prefix "\\|"
             lucy-matching-kw-regexp "\\|"
             (aref lucy-kwop-regexps lucy-max-indent-priority)))
    (cond
     ; special case for ;;
     ((and (> (point) 1) (= (preceding-char) ?\;) (= (following-char) ?\;))
      (setq in-expr nil))
     ((looking-at lucy-before-expr-prefix)
      (if (not (looking-at "(\\*")) (goto-char (match-end 0)))
      (skip-chars-forward " \t\n")
      (while (looking-at "(\\*")
        (forward-char)
        (lucy-forward-comment)
        (skip-chars-forward " \t\n"))
      (if (<= pos (point)) (setq in-expr nil))))
    (goto-char pos)
    in-expr))

(defun lucy-at-sexp-close-p ()
  (or (char-equal ?\) (following-char))
      (char-equal ?\] (following-char))
      (char-equal ?} (following-char))))

(defun lucy-find-kwop (kwop-regexp &optional min-pos)
  "Look back for a lucy keyword or operator matching KWOP-REGEXP.
Second optional argument MIN-POS bounds the search.

Ignore occurences inside literals. If found, return a list of two
values: the actual text of the keyword or operator, and a boolean
indicating whether the keyword was one we looked for explicitly
{non-nil}, or on the other hand one of the block-terminating
keywords."

  (let ((start-literal (concat "[\"" lucy-quote-char "]"))
        found kwop)
    (while (and (> (point) 1) (not found)
                (re-search-backward kwop-regexp min-pos 'move))
      (setq kwop (lucy-match-string 0))
      (cond
       ((looking-at "(\\*")
        (if (> (point) 1) (backward-char)))
       ((lucy-in-comment-p)
        (search-backward "(" min-pos 'move))
       ((looking-at start-literal))
       ((lucy-in-literal-p)
        (re-search-backward start-literal min-pos 'move))  ;ugly hack
       ((setq found t))))
    (if found
        (if (not (string-match "\\`[^|[]|[^]|]?\\'" kwop)) ;arrrrgh!!
            kwop
          (forward-char 1) "|") nil)))

;  Association list of indentation values based on governing keywords.
;
;Each element is of the form (KEYWORD OP-TYPE PRIO INDENT). OP-TYPE is
;non-nil for operator-type nodes, which affect indentation in a
;different way from keywords: subsequent lines are indented to the
;actual occurrence of an operator, but relative to the indentation of
;the line where the governing keyword occurs.

(defconst lucy-no-indent 0)

(defconst lucy-kwop-alist
  '(("automaton"            nil     6       lucy-automaton-indent)
    ("loop"            nil     6       lucy-loop-indent)
    ("class"            nil     0       lucy-class-indent)
    ("constraint"       nil     0       lucy-val-indent)
    ("sig"              nil     1       lucy-sig-indent)
    ("struct"           nil     1       lucy-struct-indent)
    ("exception"        nil     0       lucy-exception-indent)
    ("for"              nil     6       lucy-for-indent)
    ("control"              nil     6       lucy-control-indent)
    ("do"              nil     6       lucy-do-indent)    
    ("fun"              nil     3       lucy-fun-indent)
    ("function"         nil     3       lucy-function-indent)
    ("if"               nil     6       lucy-if-indent)
    ("if-else"          nil     6       lucy-if-else-indent)
    ("include"          nil     0       lucy-include-indent)
    ("inherit"          nil     0       lucy-inherit-indent)
    ("initializer"      nil     0       lucy-initializer-indent)
    ("rec"              nil     6       lucy-rec-indent)
    ("let"              nil     6       lucy-let-indent)
    ("let-in"           nil     6       lucy-let-in-indent)
    ("let-where"           nil     6       lucy-let-where-indent)
    ("match"            nil     6       lucy-match-indent)
    ("method"           nil     0       lucy-method-indent)
    ("module"           nil     0       lucy-module-indent)
    ("object"           nil     6       lucy-object-indent)
    ("of"               nil     7       lucy-of-indent)
    ("open"             nil     0       lucy-no-indent)
    ("parser"           nil     3       lucy-parser-indent)
    ("reset"           nil     6       lucy-reset-indent)
    ("try"              nil     6       lucy-try-indent)
    ("type"             nil     0       lucy-type-indent)
    ("val"              nil     0       lucy-val-indent)
    ("when"             nil     2       lucy-if-indent)
    ("::"               t       5       lucy-::-indent)
    ("@"                t       4       lucy-@-indent)
    ("^"                t       4       lucy-@-indent)
    (":="               nil     3       lucy-:=-indent)
    ("<-"               nil     3       lucy-<--indent)
    ("->"               nil     2       lucy-->-indent)
    ("\["               t       8       lucy-lb-indent)
    ("{"                t       8       lucy-lc-indent)
    ("\("               t       8       lucy-lp-indent)
    ("|"                nil     2       lucy-no-indent)
    (";;"               nil     0       lucy-no-indent))
; if-else, let-in and let-where are not keywords but idioms
; "|" is not in the regexps
; all these 4 values correspond to hard-coded names

"Association list of indentation values based on governing keywords.

Each element is of the form (KEYWORD OP-TYPE PRIO INDENT). OP-TYPE is
non-nil for operator-type nodes, which affect indentation in a
different way from keywords: subsequent lines are indented to the
actual occurrence of an operator, but relative to the indentation of
the line where the governing keyword occurs.")

;;Originally, we had lucy-kwop-regexp create these at runtime, from an
;;additional field in lucy-kwop-alist. That proved way too slow,
;;although I still can't understand why. itz

(aset lucy-kwop-regexps 0
      (concat
       "\\<\\(automaton\\|for\\|loop\\|object\\|do\\|reset\\|s\\(ig\\|truct\\)"
       "\\)\\>\\>\\|[[({]\\|;;"))
(aset lucy-kwop-regexps 1
      (concat (aref lucy-kwop-regexps 0) "\\|\\<\\(class\\|module\\)\\>"))
(aset lucy-kwop-regexps 2
      (concat
       (aref lucy-kwop-regexps 1)
       "\\|\\<\\(fun\\(ction\\)?\\|initializer\\|let\\|rec"
       "\\|m\\(atch\\|ethod\\)\\|parser\\|try\\|control\\|val\\)\\>\\|->"))
(aset lucy-kwop-regexps 3
      (concat (aref lucy-kwop-regexps 2) "\\|\\<if\\|when\\>"));\\|\\(|\\)\\(|\\)"))
(aset lucy-kwop-regexps 4
      (concat (aref lucy-kwop-regexps 3) "\\|:=\\|<-"))
(aset lucy-kwop-regexps 5
      (concat (aref lucy-kwop-regexps 4) "\\|@"))
(aset lucy-kwop-regexps 6
      (concat (aref lucy-kwop-regexps 5) "\\|::\\|\\^"))
(aset lucy-kwop-regexps 7
      (concat
       (aref lucy-kwop-regexps 0)
       "\\|\\<\\(constraint\\|exception\\|in\\(herit\\|clude\\)"
       "\\|o\\(f\\|pen\\)\\|type\\|val\\)\\>"))
(aset lucy-kwop-regexps 8
      (concat (aref lucy-kwop-regexps 6)
       "\\|\\<\\(constraint\\|exception\\|in\\(herit\\|clude\\)"
       "\\|o\\(f\\|pen\\)\\|type\\)\\>"))

(defun lucy-find-done-match ()
  (let ((unbalanced 1) (kwop t))
    (while (and (not (= 0 unbalanced)) kwop)
;;enlever le do? 
     (setq kwop (lucy-find-kwop "\\<\\(do\\(ne\\)?\\\|un\\(less\\|til\\)\\)\\>"))
      (cond
       ((not kwop))
       ((string= kwop "done") (setq unbalanced (1+ unbalanced)))
       ((or (string= kwop "until")
	    (string= kwop "unless"))
	(lucy-find-un-match))
       (t (setq unbalanced (1- unbalanced))
	  )))    
    kwop))

(defun lucy-find-end-match ()
  (let ((unbalanced 1) (kwop t))
    (while (and (not (= 0 unbalanced)) kwop)
      (setq kwop
            (lucy-find-kwop
	     (concat "\\<\\(automaton\\|end\\|for\\|loop\\|match\\|object"
		     "\\|s\\(ig\\|truct\\)\\)\\>\\>\\|;;")))
      (cond
       ((not kwop))
       ((string= kwop ";;") (setq kwop nil) (forward-line 1))
       ((string= kwop "end") (setq unbalanced (1+ unbalanced)))
       ( t (setq unbalanced (1- unbalanced)))))
    kwop
    ))

(defun lucy-find-in-match ()
  (let ((unbalanced 1) (kwop t))
    (while (and (not (= 0 unbalanced)) kwop)
      (setq kwop (lucy-find-kwop "\\<\\(in\\|let\\|end\\|where\\)\\>"))
      (let (fin beg (pos (point)))
	(end-of-line)
	(setq fin (preceding-char))
	(goto-char pos)
	(cond
	 ((not kwop))	 
	 ((string= kwop "end") (lucy-find-end-match))
	 ((string= kwop "in") (setq unbalanced (1+ unbalanced)))	
	 ((string= kwop "where") (setq unbalanced (1+ unbalanced)))
	 (t (setq unbalanced (1- unbalanced))))))
    kwop))

(defun lucy-find-where-match ()
  (let ((unbalanced 1) (kwop t))
    (while (and (not (= 0 unbalanced)) kwop)
      (setq kwop (lucy-find-kwop "\\<\\(in\\|let\\|end\\|where\\)\\>"))
      (let (fin beg (pos (point)))
	(end-of-line)
	(setq fin (preceding-char))
	(goto-char pos)
	(cond
	 ((not kwop))	 
	 ((string= kwop "end") (lucy-find-end-match))
	 ((string= kwop "in") (setq unbalanced (1+ unbalanced)))	
	 ((string= kwop "where") (setq unbalanced (1+ unbalanced)))
	 (t (setq unbalanced (1- unbalanced))))))
    kwop))

(defun lucy-find-every-match ()
  (let ((unbalanced 1) (kwop t))
    (while (and (not (= 0 unbalanced)) kwop)
      (setq kwop (lucy-find-kwop "\\<\\(reset\\|every\\|when\\)\\>"))
      (let (fin beg (pos (point)))
	(end-of-line)
	(setq fin (preceding-char))
	(goto-char pos)
	(cond
	 ((not kwop))	 
	 ((string= kwop "every") (setq unbalanced (1+ unbalanced)))	
	 ((string= kwop "when") (setq unbalanced (1+ unbalanced)))
	 (t (setq unbalanced (1- unbalanced))))))
    kwop))

(defun lucy-find-when-match ()
  (let ((unbalanced 1) (kwop t))
    (while (and (not (= 0 unbalanced)) kwop)
      (setq kwop (lucy-find-kwop "\\<\\(reset\\|every\\|when\\)\\>"))
      (let (fin beg (pos (point)))
	(end-of-line)
	(setq fin (preceding-char))
	(goto-char pos)
	(cond
	 ((not kwop))	 
	 ((string= kwop "when") (setq unbalanced (1+ unbalanced)))
	 ((string= kwop "every") (setq unbalanced (1+ unbalanced)))	
	 (t (setq unbalanced (1- unbalanced))))))
    kwop))

(defun lucy-find-un-match ()
  (let ((unbalanced 1) (kwop t))
    (while (and (not (= 0 unbalanced)) kwop)
      ;(setq kwop (lucy-find-kwop "\\<\\(until\\|done\\|do\\|unless\\)\\>\\|[^|][|][^|]"))
      (setq kwop (lucy-find-kwop "\\<\\(do\\(ne\\)?\\|end\\|un\\(less\\|til\\)\\)\\>\\|[^|][|][^|]"))
      (cond
       ((not kwop))
       ((or (string= kwop "unless")
	    (string= kwop "until")))
       ((string= kwop "done")
	(setq unbalanced (1+ unbalanced)))
        ((string= kwop "end")
 	(lucy-find-kwop "\\<\\(automaton\\)\\>"))
       ((string= kwop "|")
 	(lucy-find-kwop "\\<\\(automaton\\|match\\)\\>"))

       (t (let ((beg (point)) (prec (lucy-find-end-match)))
	    (if (not (string= prec "for"))
		(progn  (goto-char beg)  ;lucy-find-end-match a modifie la pos
		       (setq unbalanced (1- unbalanced)))
	      )))))
    kwop))

(defun lucy-find-with-match ()
  (let ((unbalanced 1) (kwop t))
    (while (and (not (= 0 unbalanced)) kwop)
      (setq kwop
        (lucy-find-kwop
         "\\<\\(with\\|try\\|control\\|m\\(atch\\|odule\\)\\|functor\\)\\>\\|[{}()]"))
      (cond
       ((not kwop))
       ((lucy-at-sexp-close-p)
        (lucy-find-paren-match (following-char)))
       ((string= kwop "with")
        (setq unbalanced (1+ unbalanced)))
       ((or (string= kwop "module")
            (string= kwop "functor")
            (string= kwop "{")
            (string= kwop "("))
        (setq unbalanced 0))
       (t (setq unbalanced (1- unbalanced)))))
    kwop))

(defun lucy-find-paren-match (close)
  (let ((unbalanced 1)
        (regexp (cond ((= close ?\)) "[()]")
                      ((= close ?\]) "[][]")
                      ((= close ?\}) "[{}]"))))
    (while (and (> unbalanced 0)
                (lucy-find-kwop regexp))
      (if (= close (following-char))
          (setq unbalanced (1+ unbalanced))
        (setq unbalanced (1- unbalanced))))))

(defun lucy-find-then-match (&optional from-else)
  (let ((bol (if from-else
                 (save-excursion
                   (progn (beginning-of-line) (point)))))
        kwop done matching-fun)
    (while (not done)
      (setq kwop
            (lucy-find-kwop
             (concat "\\<\\(continue\\|e\\(nd\\|lse\\)\\|do\\(ne\\)?\\|then"
		     "\\|un\\(less\\|til\\)\\|if\\|with\\)\\>\\|[])};]")))
      (cond
       ((not kwop) (setq done t))
       ((lucy-at-sexp-close-p)
        (lucy-find-paren-match (following-char)))
       ((string= kwop "if") (setq done t))
       ((or (string= kwop "unless")
	    (string= kwop "until"))
	(setq kwop (lucy-find-un-match))
	(setq done t))
       ((string= kwop "do")
	(let ((beg (point)) (prec (lucy-find-end-match)))
	  (goto-char beg)
	  (if (not (string= prec "for"))
	      (setq done t)
	    )))
       ((string= kwop "then")
        (if (not from-else) (setq kwop (lucy-find-then-match))))
       ((string= kwop "continue")
	(setq kwop (lucy-find-continue-match)))
       ((string= kwop ";")
	(let ((beg (point)) (prec (lucy-find-kwop "un\\(less\\|til\\)")))
	  (if (not prec)
	      (progn (goto-char beg)
		     (lucy-find-semi-match))
	    (goto-char beg))))
	  
       ((setq matching-fun (cdr-safe (assoc kwop lucy-matching-kw-alist)))
        (setq kwop (funcall matching-fun)))))
    (if (and bol (>= (point) bol))
	"if-else"
      kwop)))

(defun lucy-find-continue-match ()
  (let (; (bol (if from-else
;                  (save-excursion
;                    (progn (beginning-of-line) (point)))))
        kwop done matching-fun)
    (while (not done)
      (setq kwop
            (lucy-find-kwop
             (concat "\\<\\(continue\\|end\\|do\\(ne\\)?\\|then"
		     "\\|un\\(less\\|til\\)\\)\\>\\|[])};]")))
      (cond
       ((not kwop) (setq done t))
       ((lucy-at-sexp-close-p)
        (lucy-find-paren-match (following-char)))
       ((or (string= kwop "unless")
	    (string= kwop "until"))
	(setq kwop (lucy-find-un-match))
	(setq done t))
       ; ((string= kwop "do")
; 	(let ((beg (point)) (prec (lucy-find-end-match)))
; 	  (goto-char beg)
; 	  (if (not (string= prec "for"))
; 	      (setq done t)
; 	    )))
       ((string= kwop "then")
	(setq kwop (lucy-find-then-match)))
       ((string= kwop ";")
	(let ((beg (point)) (prec (lucy-find-kwop "un\\(less\\|til\\)")))
	  (if (not prec)
	      (progn (goto-char beg)
		     (lucy-find-semi-match))
	    (goto-char beg))))
	  
       ((setq matching-fun (cdr-safe (assoc kwop lucy-matching-kw-alist)))
        (setq kwop (funcall matching-fun)))))
;    (if (and bol (>= (point) bol))
	;"if-else"
      kwop))

(defun lucy-find-pipe-match ()
  (let ((done nil) (kwop)
        (re (concat
             "\\<\\(automaton\\|continue\\|try\\|match\\|with\\|function"
	     "\\|parser\\|type\\|e\\(nd\\|lse\\)\\|done\\|then\\|in"
	     "\\|where\\)\\>\\|[^[|]|\\|[])}]")))
    (while (not done)
      (setq kwop (lucy-find-kwop re))
      (cond
       ((not kwop) (setq done t))
       ((looking-at "[^[|]\\(|\\)")
        (goto-char (match-beginning 1))
        (setq kwop "|")
        (setq done t))
       ((lucy-at-sexp-close-p)
        (lucy-find-paren-match (following-char)))
       ((string= kwop "with")
        (setq kwop (lucy-find-with-match))
        (setq done t))
       ((string= kwop "parser")
        (if (re-search-backward "\\<with\\>" (- (point) 5) t)
            (setq kwop (lucy-find-with-match)))
        (setq done t))
       ((string= kwop "done") (lucy-find-done-match))
       ((string= kwop "end") (lucy-find-end-match))
       ((string= kwop "then") (lucy-find-then-match))
       ((string= kwop "continue") (lucy-find-continue-match))
       ((string= kwop "else") (lucy-find-else-match))
       ((string= kwop "in") (lucy-find-in-match))
       ((string= kwop "where") (lucy-find-where-match)) 
       (t (setq done t))))
    kwop))

(defun lucy-find-and-match ()
  (let ((done nil) (kwop)(v nil))
    (while (not done)
      (setq kwop (lucy-find-kwop
                  (concat "\\<\\(continue\\|rec\\|object\\|e\\(lse\\|xception\\)\\|let"
			  "\\|do\\(ne\\)?\\|t\\(hen\\|ype\\)\\|end\\|in\\|where\\)\\>")))
      (cond
       ((not kwop) (setq done t))
       ((string= kwop "end") (lucy-find-end-match))
       ((string= kwop "done") (lucy-find-done-match))
       ((string= kwop "else") (lucy-find-else-match))
       ((string= kwop "rec")
	(let ((beg (point)))
	  (goto-char (- (point) 4))
	  (if (looking-at "let")
	      (progn (setq kwop "let")
		     (setq done t)
		     (goto-char beg))
	    (goto-char beg)
	    (setq done t))))
       ((string= kwop "in") (lucy-find-in-match))
       ((string= kwop "where") (lucy-find-where-match))
       ((string= kwop "do") ; (let ((beg (point)) (prec (lucy-find-end-match)))
; ; 				   (goto-char beg)
; ; 				   (if (string= prec "for")
; ; 				       nil
; ; 				     t)))
	     ;(not v))
	(if v
	    (setq v nil)
	  (setq done t)))
 	     ; (let ((beg (point)) (prec (lucy-find-kwop "\\<\\(automaton\\)\\>")))
;  	       (if prec
;  		   t
;  		 (goto-char beg)
;  		 nil)))
;	     (setq done t)
; 	)
       ((string= kwop "then")
	(let ((beg (point)) (prec (lucy-find-then-match)))
	  (if prec 
	      prec
	    (goto-char beg)  ;lucy-find-done-match a modifie la pos
	    (setq v t))))
       ((string= kwop "continue")
	(setq v t))
       (t (setq done t))))
    kwop))

(defun lucy-find-else-match ()
  (lucy-find-then-match t))

(defun lucy-find-semi-match ()
  (lucy-find-kwop-skipping-blocks 2))

(defun lucy-find-comma-match ()
  (lucy-find-kwop-skipping-blocks 3))

(defun lucy-find-kwop-skipping-blocks (prio)
  "Look back for a lucy keyword matching lucy-kwop-regexps [PRIO].

 Skip nested blocks."

  (let ((done nil) (kwop nil) (matching-fun)(pos (point))
        (kwop-list (aref lucy-kwop-regexps prio)))
    (while (not done)
      (setq kwop (lucy-find-kwop
                  (concat lucy-matching-kw-regexp
                          (cond ((> prio 3) "\\|[])},;]\\|")
                                ((> prio 2) "\\|[])};]\\|")
                                (t "\\|[])}]\\|"))
                          kwop-list)))
      (cond
       ((not kwop) (setq done t))
       ((looking-at "rec")  
	(let ((beg (point)))
	  (goto-char (- (point) 4))
	  (if (looking-at "let")
	      (goto-char beg)
	    (goto-char beg)   ; if rec without let
	    (setq done t))))
	    
       ((lucy-at-sexp-close-p)  
        (lucy-find-paren-match (following-char)))
       ((or (string= kwop ";;")
	    (and (string= kwop ";") (= (preceding-char) ?\;)))
	(forward-line 1)
        (setq kwop ";;")
        (setq done t))
       ((and (>= prio 2) (string= kwop "|")) (setq done t))
       ((string= kwop "end") (lucy-find-end-match))
       ((string= kwop "done") (lucy-find-done-match))
       ((string= kwop "where")
        (cond ((and (lucy-find-where-match) (>= prio 2))
               (setq kwop "let-where")
               (setq done t))))
       ((string= kwop "in") 
	(let ((mot (lucy-find-in-match)))
	  (cond ((and mot (>= prio 2))
		 (cond ((string= "let" mot)
			(setq kwop "let-in"))
		       ((string= "signal" mot)
			(setq kwop "signal-in"))
		       )
		 (setq done t)))))
       ((and (string= kwop "parser") (>= prio 2)
             (re-search-backward "\\<with\\>" (- (point) 5) t))
        (setq kwop (lucy-find-with-match))
        (setq done t))
       ((and (if (looking-at "do[^n]")
		 (let ((beg (point)) (prec (lucy-find-end-match)))
		   (goto-char beg)
		   (if (string= prec "for")
		       prec
		     nil))
	       t)
	     (if (looking-at "when")
		 (let ((beg (point)) (prec (lucy-find-when-match)))
		   (goto-char beg)
		   (if prec
		       t    ;if reset-when form
		     nil))
	       t)
	     (setq matching-fun (cdr-safe (assoc kwop lucy-matching-kw-alist))))
	(if (and (or (looking-at "when") (looking-at "every"))
		 (= (lucy-verif-boucle pos) 1)) ;if not first statement
	    (lucy-find-every-match)
	  (setq kwop (funcall matching-fun))
	  (if (looking-at kwop-list) (setq done t))
	))
       ((and (looking-at "when")
	    (< prio 3))	
	(setq kwop nil))
       (t 
	(let* ((kwop-info (assoc kwop lucy-kwop-alist))
	       (is-op (and (nth 1 kwop-info)
					; check that we are not at beginning of line
			   (let ((pos (point)) bti)
                               (back-to-indentation)
                               (setq bti (point))
                               (goto-char pos)
                               (< bti pos)))))
            (if (and is-op (looking-at
                            (concat (regexp-quote kwop)
                                    "|?[ \t]*\\(\n\\|(\\*\\)")))
                (setq kwop-list
                      (aref lucy-kwop-regexps (nth 2 kwop-info)))
              (setq done t))))))
    kwop))

(defun lucy-compute-basic-indent (prio)
  "Compute indent of current lucy line, ignoring leading keywords.

Find the `governing node' for current line. Compute desired
indentation based on the node and the indentation alists.
Assumes point is exactly at line indentation.
Does not preserve point."

  (let* (in-expr
	 (kwop (cond
                ((looking-at ";;")
                 (beginning-of-line 1))
                ((looking-at "|\\([^]|]\\|\\'\\)")
                 (lucy-find-pipe-match))
		((and (looking-at "rec")
		      (let ((beg (point)))
			(goto-char (- (point) 4))
			(if (looking-at "let")
			    (progn (goto-char beg)
				   nil)
			  (goto-char beg) ; if rec without let
			  (setq in-expr t)
			  t)))
		 (lucy-find-kwop-skipping-blocks 6))
		((and (looking-at lucy-phrase-start-keywords)
                      (lucy-in-expr-p))
		 ; (let ((beg (point)) (kw (lucy-find-end-match)))
;  		   (goto-char beg)
;  		   (if (and (looking-at "let") (string= kw "automaton"))
;  		       (lucy-find-kwop "state")
;  		     kw)))
		 )
		((looking-at "do[^n]")
		 (let ((beg (point)) (prec (lucy-find-end-match)))
		   (if (string= prec "for") 
		       prec
		     (goto-char beg)  ;lucy-find-done-match a modifie la pos
		     (lucy-find-kwop-skipping-blocks 6))))
		((looking-at "when") 
		 (let ((beg (point)) (prec (lucy-find-when-match)))
		   (if prec 
		       prec		;if not do-when form     
		     (goto-char beg) 
		     (lucy-find-kwop-skipping-blocks 2))))
		((and (looking-at lucy-matching-kw-regexp)
                      (assoc (lucy-match-string 0) lucy-matching-kw-alist))
                 (funcall (cdr-safe (assoc (lucy-match-string 0)
					   lucy-matching-kw-alist))))
                ((looking-at
                  (aref lucy-kwop-regexps lucy-max-indent-priority))
                 (let* ((kwop (lucy-match-string 0))
                        (kwop-info (assoc kwop lucy-kwop-alist))
                        (prio (if kwop-info (nth 2 kwop-info)
                                lucy-max-indent-priority)))
		   (if (and (looking-at (aref lucy-kwop-regexps 0))
                            (not (looking-at "object"))
			    (not (looking-at "||"))
			    (lucy-in-expr-p))
		       (setq in-expr t))
		   (lucy-find-kwop-skipping-blocks prio)))
                (t 
		 (if (and (= prio lucy-max-indent-priority) (lucy-in-expr-p))
		     (setq in-expr t))
		 (lucy-find-kwop-skipping-blocks prio))))
         (kwop-info (assoc kwop lucy-kwop-alist))
         (indent-diff
          (cond
           ((not kwop-info) (beginning-of-line 1) 0)
           ((looking-at "[[({][|<]?[ \t]*")
            (length (lucy-match-string 0)))
           ((nth 1 kwop-info) (symbol-value (nth 3 kwop-info)))
           (t 
            (let ((pos (point)))
              (back-to-indentation)
	      (if (looking-at "\\<do\\|rec\\|and\\|match\\|with\\>") (goto-char pos))
              (- (symbol-value (nth 3 kwop-info))
                 (if (and (looking-at "|") (not (looking-at "||"))) 
		     (progn  lucy-|-extra-indent) 
		    0))))))
         (extra (if in-expr lucy-apply-extra-indent 0)))
    (+ indent-diff extra (current-column))))

(defun lucy-verif-boucle (pos)
  "Look after the word to see if the line to indent is the first 
statement after the keyword (return 0) or not (return 1)"

   (let ((pos2 (+ (point) 5)))
     (goto-char pos2)
     (let ((d (skip-chars-forward " \t\n" pos)))       
	       (if (= d (- pos pos2))
		   (progn (goto-char (- pos2 5)) 0)
		 (goto-char (- pos2 5))
		 1))))

(defconst lucy-leading-kwops-regexp
  (concat
   "\\<\\(and\\|continue\\|do\\(ne\\)?\\|e\\(lse\\|nd\\|very\\)\\|in"
   "\\|t\\(hen\\|o\\)\\|un\\(less\\|til\\)\\|w\\(he\\(n\\|re\\)\\|ith\\)"
   "\\)\\>\\|[]|})]")

  "Regexp matching lucy keywords which need special indentation.")

(defconst lucy-leading-kwops-alist
  '(("and" lucy-and-extra-indent 2)
    ("continue" lucy-continue-extra-indent 0)
    ("do" lucy-do-extra-indent 0)
    ("done" lucy-done-extra-indent 0)
    ("else" lucy-else-extra-indent 3)
    ("end" lucy-end-extra-indent 0)
    ("every" lucy-every-extra-indent 2)
    ("in" lucy-in-extra-indent 2)
    ("then" lucy-then-extra-indent 3)
    ("to" lucy-to-extra-indent 0)
    ("until" lucy-until-extra-indent 2)
    ("unless" lucy-unless-extra-indent 2)
    ("when" lucy-when-extra-indent 2)
    ("where" lucy-where-extra-indent 2)
    ("with" lucy-with-extra-indent 2)
    ("|" lucy-|-extra-indent 2)
    ("]" lucy-rb-extra-indent 0)
    ("}" lucy-rc-extra-indent 0)
    (")" lucy-rp-extra-indent 0))

  "Association list of special lucy keyword indent values.

Each member is of the form (KEYWORD EXTRA-INDENT PRIO) where
EXTRA-INDENT is the variable holding extra indentation amount for
KEYWORD (usually negative) and PRIO is upper bound on priority of
matching nodes to determine KEYWORD's final indentation.")

(defun lucy-compute-final-indent ()
  (save-excursion
    (back-to-indentation)
    (cond
     ((and (bolp) (looking-at comment-start-skip)) (current-column))
     ((lucy-in-comment-p)
      (let ((closing (looking-at "\\*)"))
            (comment-mark (looking-at "\\*")))
        (lucy-backward-comment)
        (looking-at comment-start-skip)
        (+ (current-column)
           (cond
            (closing 1)
            (comment-mark 1)
            (t lucy-comment-indent)))))
     ((looking-at "do[^n]")
      (let ((beg (point)) (prec (lucy-find-end-match)))
	(if (string= prec "for")
	    (progn	
	      (goto-char beg)
	      (let ((basic (lucy-compute-basic-indent 0)))
		(max 0 (current-column))))
	  (goto-char beg)  ;lucy-find-end-match a modife la pos
	  (let ((basic (lucy-compute-basic-indent lucy-max-indent-priority)))
	    (max 0 basic)))))
     ;; ((looking-at "and")
;;       (let ((beg (point)) (prec (lucy-find-end-match)))
;; 	(if (string= prec "automaton")
;; 	    (progn	
;; 	      (goto-char beg)
;; 	      (let ((basic (lucy-compute-basic-indent 0)))
;; 		(max 0 (current-column))))
;; 	  (goto-char beg)  ;lucy-find-end-match a modife la pos
;; 	  (let ((basic (lucy-compute-basic-indent lucy-max-indent-priority)))
;; 	    (max 0 basic)))))
     ((looking-at "when")
      (let ((beg (point)) (prec (lucy-find-when-match)))
	(if prec 
	    (progn	      
	      (goto-char beg)
	      (let ((basic (lucy-compute-basic-indent 0)))
		(max 0 (current-column))))	  
	  (goto-char beg)  ;lucy-find-when-match a modife la pos
	  (let ((basic (lucy-compute-basic-indent lucy-max-indent-priority)))
	    (max 0 basic)))))
     (t (let* ((leading (and (looking-at lucy-leading-kwops-regexp)
					(not (looking-at "||"))))
               (assoc-val (if leading (assoc (lucy-match-string 0)
                                             lucy-leading-kwops-alist)))
               (extra (if leading (symbol-value (nth 1 assoc-val)) 0))
               (prio (if leading (nth 2 assoc-val)
                       lucy-max-indent-priority))
               (basic (lucy-compute-basic-indent prio)))
	  (max 0 (if extra (+ extra basic) (current-column))))))))


(defun lucy-split-string ()
  "Called whenever a line is broken inside a lucy string literal."
  (insert-before-markers "\"^\"")
  (backward-char 1))

(defadvice indent-new-comment-line (around
                                    lucy-indent-new-comment-line
                                    activate)

  "Handle multi-line strings in lucy mode."

;this advice doesn't make sense in other modes. I wish there were a
;cleaner way to do this: I haven't found one.

  (let ((hooked (and (eq major-mode 'lucy-mode) (lucy-in-literal-p)))
        (split-mark))
    (if (not hooked) nil
      (setq split-mark (set-marker (make-marker) (point)))
      (lucy-split-string))
    ad-do-it
    (if (not hooked) nil
      (goto-char split-mark)
      (set-marker split-mark nil))))

(defadvice newline-and-indent (around
                               lucy-newline-and-indent
                               activate)

  "Handle multi-line strings in lucy mode."

    (let ((hooked (and (eq major-mode 'lucy-mode) (lucy-in-literal-p)))
        (split-mark))
    (if (not hooked) nil
      (setq split-mark (set-marker (make-marker) (point)))
      (lucy-split-string))
    ad-do-it
    (if (not hooked) nil
      (goto-char split-mark)
      (set-marker split-mark nil))))

(defun lucy-electric-pipe ()
  "If inserting a | or } operator at beginning of line, reindent the line.

Unfortunately there is a situation where this mechanism gets
confused. It's when | is the first character of a |] sequence. This is
a misfeature of lucy syntax and cannot be fixed, however, as a
workaround, the electric ] inserts | itself if the matching [ is
followed by |."

  (interactive "*")
  (let ((electric (and lucy-electric-indent
                       (lucy-in-indentation)
                       (not (lucy-in-comment-p)))))
    (self-insert-command 1)
    (if electric (save-excursion (lucy-indent-command)))))

(defun lucy-electric-rb ()
  "If inserting a ] operator at beginning of line, reindent the line.

Also, if the matching [ is followed by a | and this ] is not preceded
by |, insert one."

  (interactive "*")
  (let* ((prec (preceding-char))
         (use-pipe (and lucy-electric-close-vector
                        (not (lucy-in-comment-p))
                        (not (lucy-in-literal-p))
                        (or (not (numberp prec))
                            (not (char-equal ?| prec)))))
         (electric (and lucy-electric-indent
                        (lucy-in-indentation)
                        (not (lucy-in-comment-p)))))
    (self-insert-command 1)
    (if electric (save-excursion (lucy-indent-command)))
    (if (and use-pipe
             (save-excursion
               (condition-case nil
                   (prog2
                       (backward-list 1)
                       (looking-at "\\[|"))
                 (error ""))))
        (save-excursion
          (backward-char 1)
          (insert "|")))))

(defun lucy-abbrev-hook ()
  "If inserting a leading keyword at beginning of line, reindent the line."
  ;itz unfortunately we need a special case
  (if (and (not (lucy-in-comment-p)) (not (= last-command-char ?_)))
      (let* ((bol (save-excursion (beginning-of-line) (point)))
             (kw (save-excursion
                   (and (re-search-backward "^[ \t]*\\(\\sw+\\)\\=" bol t)
                        (lucy-match-string 1)))))
        (if kw
            (let ((indent (save-excursion
                            (goto-char (match-beginning 1))
                            (lucy-indent-command)
                            (current-column)))
                  (abbrev-correct (if (= last-command-char ?\ ) 1 0)))
              (indent-to (- indent
                            (or
                             (symbol-value
                              (nth 1
                                   (assoc kw lucy-leading-kwops-alist)))
                             0)
                            abbrev-correct)))))))

; (defun lucy-indent-phrase ()
;   (interactive "*")
;   (let ((bounds (lucy-mark-phrase)))
;     (indent-region (car bounds) (cdr bounds) nil)))

;;; Additional commands by Didier to report errors in toplevel mode

(defun lucy-skip-blank-forward ()
  (if (looking-at "[ \t\n]*\\((\\*\\([^*]\\|[^(]\\*[^)]\\)*\\*)[ \t\n]*\\)*")
      (goto-char (match-end 0))))

;; to mark phrases, so that repeated calls will take several of them
;; knows little about Lucy appart literals and comments, so it should work
;; with other dialects as long as ;; marks the end of phrase. 

(defun lucy-indent-phrase (arg)
  "Indent current phrase
with prefix arg, indent that many phrases starting with the current phrase."
  (interactive "p")
  (save-excursion
    (let ((beg (lucy-find-phrase)))
    (while (progn (setq arg (- arg 1)) (> arg 0)) (lucy-find-phrase))
    (indent-region beg (point) nil))))

(defun lucy-indent-buffer ()
  (interactive)
  (indent-region (point-min) (point-max) nil))

(defun lucy-backward-to-less-indent (&optional n)
  "Move cursor back  N lines with less or same indentation."
  (interactive "p")
  (beginning-of-line 1)
  (if (< n 0) (lucy-forward-to-less-indent (- n))
    (while (> n 0)
      (let ((i (current-indentation)))
        (forward-line -1)
        (while (or (> (current-indentation) i)
                   (lucy-in-comment-p)
                   (looking-at
                    (concat "[ \t]*\\(\n\\|" comment-start-skip "\\)")))
          (forward-line -1)))
      (setq n (1- n))))
  (back-to-indentation))

(defun lucy-forward-to-less-indent (&optional n)
  "Move cursor back N lines with less or same indentation."
  (interactive "p")
  (beginning-of-line 1)
  (if (< n 0) (lucy-backward-to-less-indent (- n))
    (while (> n 0)
      (let ((i (current-indentation)))
        (forward-line 1)
        (while (or (> (current-indentation) i)
                   (lucy-in-comment-p)
                   (looking-at
                    (concat "[ \t]*\\(\n\\|" comment-start-skip "\\)")))
          (forward-line 1)))
      (setq n (1- n))))
  (back-to-indentation))

(defun lucy-insert-automaton-form ()
  "Inserts a nicely formatted automaton-end form, leaving a mark after end."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)) (i (+ lucy-automaton-indent c)))
    (indent-line-to c)
    (insert "automaton ")
     (indent-line-to c)
     (push-mark)
     (forward-line -1)
     (indent-line-to i)
     (end-of-line)))
    
(defun lucy-insert-loop-form ()
  "Inserts a nicely formatted loop-end form, leaving a mark after end."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)) (i (+ lucy-loop-indent c)))
    (insert "loop\n\nend")
    (push-mark)
    (indent-line-to c)
    (forward-line -1)
    (indent-line-to i)))

(defun lucy-insert-for-form ()
  "Inserts a nicely formatted for-do-end form, leaving a mark after do."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)) (i (+ lucy-for-indent c)))
    (insert "for  do\n\nend")
    (push-mark)
    (indent-line-to c)
    (forward-line -1)
    (indent-line-to i)
    (push-mark)
    (beginning-of-line 1)
    (backward-char 4)))

(defun lucy-insert-if-form ()
  "Insert nicely formatted if-then-else form leaving mark after then, else."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)) (i (+ lucy-if-indent c)))
    (insert "if\n\nthen\n\nelse\n")
    (indent-line-to i)
    (push-mark)
    (forward-line -1)
    (indent-line-to c)
    (forward-line -1)
    (indent-line-to i)
    (push-mark)
    (forward-line -1)
    (indent-line-to c)
    (forward-line -1)
    (indent-line-to i)))

(defun lucy-insert-match-form ()
  "Insert nicely formatted match-with form leaving mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)) (i (+ lucy-match-indent c)))
    (insert "match\n\nwith\n")
    (indent-line-to i)
    (push-mark)
    (forward-line -1)
    (indent-line-to c)
    (forward-line -1)
    (indent-line-to i)))

(defun lucy-insert-let-form ()
  "Insert nicely formatted let-in form leaving mark after in."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)))
    (insert "let  in\n")
    (indent-line-to c)
    (push-mark)
    (forward-line -1)
    (forward-char (+ c 4))))

(defun lucy-insert-try-form ()
  "Insert nicely formatted try-with form leaving mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)) (i (+ lucy-try-indent c)))
    (insert "try\n\nwith\n")
    (indent-line-to i)
    (push-mark)
    (forward-line -1)
    (indent-line-to c)
    (forward-line -1)
    (indent-line-to i)))

(defun lucy-insert-control-form ()
  "Insert nicely formatted control-with form leaving mark after with."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)) (i (+ lucy-control-indent c)))
    (insert "control\n\nwith\n")
    (indent-line-to i)
    (push-mark)
    (forward-line -1)
    (indent-line-to c)
    (forward-line -1)
    (indent-line-to i)))

(defun lucy-insert-reset-form ()
  "Insert nicely formatted do-until form leaving mark after until."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)) (i (+ lucy-do-indent c)))
    (indent-line-to c)
    (insert "reset \n\nevery ")
    (indent-line-to c)
    (push-mark)
    (forward-line -1)
    (end-of-line)))

(defun lucy-insert-do-until-form ()
  "Insert nicely formatted do-until form leaving mark after do, done."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)) (i (+ lucy-do-indent c)))
    (indent-line-to c)
    (insert "do \nuntil  then ")
    (indent-line-to c)
    (end-of-line)
    (backward-char 6)
    (push-mark)
    (forward-line -1)
    (end-of-line)))

(defun lucy-insert-do-unless-form ()
  "Insert nicely formatted do-unless form leaving mark after unless."
  (interactive "*")
  (let ((prec (preceding-char)))
    (if (and (numberp prec) (not (char-equal ?\  (char-syntax prec))))
        (insert " ")))
  (let* ((c (current-indentation)) (i (+ lucy-do-indent c)))
    (indent-line-to c)
    (insert "do \nunless  then ")
    (indent-line-to c)
    (forward-char 7)
    (push-mark)
    (forward-line -1)
    (end-of-line)))

(autoload 'run-lucy "inf-lucy" "Run an inferior Lucy process." t)

(autoload 'lucy-types-show-type "lucy-types"
  "Show the type of expression or pattern at point." t)
(autoload 'lucy-types-explore "lucy-types"
  "Explore type annotations by mouse dragging." t)

(autoload 'lucy-help "lucy-help"
  "Show documentation for qualilifed Lucy identifier." t)
(autoload 'lucy-complete "lucy-help"
  "Does completion for documented qualified Lucy identifier." t)
(autoload 'lucy-open-module "lucy-help"
  "Add module in documentation search path." t)
(autoload 'lucy-close-module "lucy-help"
  "Remove module from documentation search path." t)
(autoload 'lucy-add-path "lucy-help"
  "Add search path for documentation." t)

;;; lucy.el ends here

(provide 'lucy)
