(* $Id: simu.ls,v 1.2 2006/12/29 17:19:31 pouzet Exp $ *)
open Graphics
open Chrono
open Heater

(** we add some stub code to prepare the simulation                      *)
(* we do something simple here: the environment is                       *)
(* given by the user. In a more realistic simulation, the                *)
(* behavior of the heater is described by a synchronous code + noise     *)
(* and making a closed loop of the two (+ the user to give extra inputs) *)
(*                                                                       *)
(*                                       noise                           *)
(*                                        |                              *)
(*                                        \/                             *)
(*                  -----------        ---------                         *)
(*                  |         |        |       |                         *)
(* ------ +/- ----> | Control | -----> | Env   |                         *)
(*        /\        |         |        |       |                         *)
(*         |        ----------         ---------                         *)
(*         --------------------------------|                             *)
(*                                                                       *)
(*                                                                       *)

type button = Up | Down

(* transforming a two-position button into an integer value *)
let node button_to_integer init step button = temp where
  rec last temp = init
  and present
        button(v) ->
	  do match v with
	       Up -> do temp = last temp + step done
	     | Down -> do temp = last temp - step done
	     end
	  done
      end

(* the interface is done through the keybord *)
let node environment key =
  let 
    (* build the various input/output events from the environment *)
    present
      key(v) ->
	do match v with
	     'P' -> do emit expected_temp = Up done
	   | 'M' -> do emit expected_temp = Down done
	   | 'E' -> do emit actual_temp = Up done
	   | 'Z' -> do emit actual_temp = Down done
	   | 'R' -> do emit restart = () done
	   | 'L' -> do emit on_light = () done
	   | _ -> do done
	   end
	done
    end in
  let
      expected_temp = button_to_integer 20 1 expected_temp in
  let
      actual_temp = button_to_integer 20 1 actual_temp in
  let
      on_light = ?(on_light) in
  let 
      restart = ?(restart) in

  expected_temp, actual_temp, on_light, restart

(** outputs *)
(*  the function [print] emits the outputs on stdout *)
let print open_light open_gas ok nok expected_temp actual_temp 
    restart on_light =
  print_string "open_light = ";
  print_string (if open_light then "on" else "off");
  print_string " open_gas = ";
  print_string (if open_gas then "on" else "off");
  print_string (if ok then " ok" else if nok then " nok" else "");
  print_string " expected_temp = ";
  print_int expected_temp;
  print_string " actual_temp = ";
  print_int actual_temp;
  print_string " restart = ";
  print_string (if restart then "on" else "off");
  print_string " on_light = ";
  print_string (if on_light then "on" else "off");
  print_newline ();
  flush stdout

(* outputs for sim2chro *)
let node chronogram (static oc)
    open_light open_gas ok nok expected_temp 
    actual_temp restart on_light =
  let initialisation, running = Chrono.sim2chro oc in
  let static program = "Heater" in
  let static def_inputs = 
    [ "expected_temp", Tint;
      "actual_temp", Tint;
      "restart", Tbool;
      "on_light", Tbool ] in
  let static def_outputs =
    [ "open_light", Tbool;
      "open_gas", Tbool;
      "ok", Tbool;
      "nok", Tbool ] in
  let inputs =
    [ Vint(expected_temp); Vint(actual_temp); Vbool(restart);
      Vbool(on_light)] in
  let outputs =
    [ Vbool(open_light); Vbool(open_gas); Vbool(ok); Vbool(nok) ] in
    
  match true -> false with
      true -> initialisation program def_inputs def_outputs
    | false -> ()
  end;
  running (inputs, outputs)

let keyboard status = key where
  match status.keypressed with
      true -> do emit key = status.key done
    | false -> do done
  end

(* the base time is 1/100 of a second *)
let node system (static oc) status =
  let key = keyboard status in
  (* the external environment is simulated by the user *)
  let expected_temp, actual_temp, on_light, restart = environment key in
  
  (* put controller in the loop *)
  let open_light, open_gas, ok, nok = 
    controller true restart expected_temp actual_temp on_light in
  
  (* show the chronogram *)
  chronogram oc open_light open_gas ok nok 
    expected_temp actual_temp restart on_light
  
